from __future__ import annotations

import math

import pytest

from tests.helpers import build_args
from visisipy.models import EyeModel
from visisipy.types import SampleSize

pytestmark = [pytest.mark.needs_opticstudio]


class MockPupil:
    def __init__(self):
        self.semi_diameter_history = []
        self.semi_diameter = 0.5

    @property
    def semi_diameter(self):
        return self._semi_diameter

    @semi_diameter.setter
    def semi_diameter(self, value):
        self.semi_diameter_history.append(value)
        self._semi_diameter = value


class MockOpticstudioModel:
    def __init__(self):
        self._pupil = MockPupil()

    @property
    def pupil(self):
        return self._pupil


@pytest.fixture
def mock_update_pupil(opticstudio_backend, monkeypatch):
    monkeypatch.setattr(opticstudio_backend, "aperture_history", [], raising=False)

    @classmethod
    def update_pupil(cls, pupil_diameter):
        cls.aperture_history.append(pupil_diameter)

    monkeypatch.setattr(opticstudio_backend, "update_pupil", update_pupil)


class TestRefractionAnalysis:
    @pytest.mark.parametrize(
        "field_coordinate,wavelength,sampling,pupil_diameter,field_type,use_higher_order_aberrations",
        [
            (None, None, None, None, None, None),
            ((0, 0), 0.543, None, None, None, None),
            ((1, 1), 0.632, 64, None, None, None),
            ((math.pi, math.tau), 0.543, "64x64", 0.5, None, None),
            ((0, 0), 0.543, "64x64", 0.5, "angle", None),
            ((1, 1), 0.632, SampleSize(32), 0.5, "object_height", None),
            ((0, 0), 0.543, "32x32", 0.5, "angle", True),
            ((1, 1), 0.632, "64x64", 0.5, "object_height", False),
        ],
    )
    def test_refraction(
        self,
        field_coordinate,
        wavelength,
        sampling,
        pupil_diameter,
        field_type,
        use_higher_order_aberrations,
        opticstudio_analysis,
        opticstudio_backend,
        monkeypatch,
    ):
        opticstudio_backend.build_model(
            EyeModel(), object_distance=10 if field_type == "object_height" else float("inf")
        )

        args = build_args(
            field_coordinate=field_coordinate,
            wavelength=wavelength,
            sampling=sampling,
            pupil_diameter=pupil_diameter,
            field_type=field_type,
            use_higher_order_aberrations=use_higher_order_aberrations,
            non_null_defaults={"sampling", "field_type", "use_higher_order_aberrations"},
        )

        assert opticstudio_analysis.refraction(**args)

    @pytest.mark.parametrize(
        "pupil_diameter,change_pupil_diameter",
        [
            (None, False),
            (1.234, True),
        ],
    )
    def test_change_pupil(
        self, opticstudio_backend, mock_update_pupil, opticstudio_analysis, pupil_diameter, change_pupil_diameter
    ):
        opticstudio_backend.update_settings(aperture_type="float_by_stop_size", aperture_value=1.0)

        assert opticstudio_backend.aperture_history == []

        opticstudio_analysis.refraction(pupil_diameter=pupil_diameter)

        if change_pupil_diameter:
            assert opticstudio_backend.aperture_history == [pupil_diameter, 1.0]
        else:
            assert opticstudio_backend.aperture_history == []

    @pytest.mark.parametrize(
        "aperture_type,pupil_diameter",
        [
            ("entrance_pupil_diameter", 1.0),
            ("image_f_number", 2.0),
            ("object_numeric_aperture", 0.1),
        ],
    )
    def test_change_pupil_warns_aperture(
        self, pupil_diameter, aperture_type, opticstudio_backend, opticstudio_analysis
    ):
        opticstudio_backend.build_model(
            EyeModel(), object_distance=10 if aperture_type == "object_numeric_aperture" else float("inf")
        )
        opticstudio_backend.update_settings(aperture_type=aperture_type)

        with pytest.warns(
            UserWarning, match="When updating the pupil size for aperture types other than 'float_by_stop_size'"
        ):
            opticstudio_analysis.refraction(pupil_diameter=pupil_diameter)



"""Spherical equivalent of refraction analysis for Optiland."""

from __future__ import annotations

from typing import TYPE_CHECKING
from warnings import warn

from visisipy.analysis.refraction import zernike_data_to_refraction
from visisipy.optiland.analysis.zernike_coefficients import (
    zernike_standard_coefficients,
)
from visisipy.types import SampleSize

if TYPE_CHECKING:
    from optiland.wavefront import ZernikeOPD

    from visisipy.backend import FieldCoordinate, FieldType
    from visisipy.optiland.backend import OptilandBackend
    from visisipy.refraction import FourierPowerVectorRefraction

__all__ = ("refraction",)


def refraction(
    backend: type[OptilandBackend],
    field_coordinate: FieldCoordinate | None = None,
    wavelength: float | None = None,
    sampling: SampleSize | str | int = 64,
    pupil_diameter: float | None = None,
    field_type: FieldType = "angle",
    *,
    use_higher_order_aberrations: bool = True,
) -> tuple[FourierPowerVectorRefraction, ZernikeOPD]:
    """Calculates the ocular refraction.

    The ocular refraction is calculated from Zernike standard coefficients and represented in Fourier power
    vector form.

    Parameters
    ----------
    backend : type[OptilandBackend]
        Reference to the OptilandBackend backend.
    field_coordinate : tuple[float, float], optional
        The field coordinate for the Zernike calculation. When `None`, the first field in Optiland is used.
        Defaults to `None`.
    wavelength : float, optional
        The wavelength for the Zernike calculation. When `None`, the first wavelength in Optiland is used.
        Defaults to `None`.
    sampling : SampleSize | str | int, optional
        The sampling for the Zernike calculation. Defaults to 64.
    pupil_diameter : float, optional
        The diameter of the pupil for the refraction calculation. Defaults to the pupil diameter configured in the
        backend. If the aperture type is "float_by_stop_size", the value is interpreted as the pupil diameter.
        For other aperture types, it is interpreted as the aperture value.
    field_type : Literal["angle", "object_height"], optional
        The type of field to be used when setting the field coordinate. This parameter is only used when
        `field_coordinate` is specified. Defaults to "angle".
    use_higher_order_aberrations : bool, optional
        If `True`, higher-order aberrations are used in the calculation. Defaults to `True`.

    Returns
    -------
    FourierPowerVectorRefraction
        The ocular refraction in Fourier power vector form.
    """
    old_aperture = None
    if pupil_diameter is not None:
        if backend.get_setting("aperture_type") != "float_by_stop_size":
            message = (
                "When updating the pupil size for aperture types other than 'float_by_stop_size', "
                "the pupil_diameter parameter changes the aperture value instead of the pupil size."
            )
            warn(message, UserWarning, stacklevel=2)

        old_aperture = backend.get_optic().aperture
        backend.update_pupil(pupil_diameter)

    zernike_coefficients, zernike_opd = zernike_standard_coefficients(
        backend=backend,
        field_coordinate=field_coordinate,
        wavelength=wavelength,
        field_type=field_type,
        sampling=SampleSize(sampling),
    )

    exit_pupil_semi_diameter = backend.get_optic().paraxial.XPD() / 2

    if old_aperture is not None:
        backend.update_pupil(old_aperture.value)

    if wavelength is None:
        wavelength = backend.get_wavelengths()[0]

    fourier_refraction = zernike_data_to_refraction(
        zernike_coefficients,
        exit_pupil_semi_diameter=exit_pupil_semi_diameter,
        wavelength=wavelength,
        use_higher_order_aberrations=use_higher_order_aberrations,
    )

    return fourier_refraction, zernike_opd
